#include "MeshLoader.h"
#include "HashCustom.h"
#include <fstream>
#include "Exceptions.h"
#include <ranges>
#include <algorithm>
#include <vector>
#include <iterator>
#include <iostream>



const std::vector<Node>& MeshLoader::Get_Node() const {
    return nodes;
}

const std::vector<FiniteElement>& MeshLoader::Get_Finite_Elements() const {
    return finite_elements;
}

const std::vector<BoundaryFiniteElement>& MeshLoader::Get_Boundary() const {
    return boundary;
}

std::set<int> MeshLoader::Get_Nodes_By_Surface(int t_surface_id) const
{
	std::set<int> res;
	auto first_to_find = boundary.begin();
	auto last_to_find = boundary.end();

	while (first_to_find != last_to_find) {
		auto found_boundary = std::find_if(first_to_find, last_to_find,
			[t_surface_id](const BoundaryFiniteElement& cur_boundary) {
				return cur_boundary.surface_id == t_surface_id;
			});

		first_to_find = found_boundary;
		if (found_boundary != last_to_find) {
			for (const auto& node_in_boundary : found_boundary->nodes_id) res.insert(node_in_boundary);
			++first_to_find;
		}
	}
	return res;
}

std::vector<int> MeshLoader::Get_Elements_By_ID(int t_node_id1, int t_node_id2, int t_node_id3) const
{
	std::vector<int> res;
	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_node_id1, t_node_id2, t_node_id3](const FiniteElement& elem) {
				return std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id1) != elem.nodes_id.end()
	&& std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id2) != elem.nodes_id.end()
	&& std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id3) != elem.nodes_id.end();
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

std::vector<int> MeshLoader::Get_Elements_By_Edge(int t_node_id1, int t_node_id2) const
{
	std::vector<int> res;

	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_node_id1, t_node_id2](const FiniteElement& elem) {
				return std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id1) != elem.nodes_id.end()
					&& std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id2) != elem.nodes_id.end();
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

std::vector<int> MeshLoader::Get_Elements_By_Material(int t_material_id) const
{
	std::vector<int> res;

	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_material_id](const FiniteElement& elem) {
				return elem.material_id == t_material_id;
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

std::vector<int> MeshLoader::Get_Boundary_By_Surface(int t_surface_id) const
{
	std::vector<int> res;

	auto first_to_find = boundary.begin();
	auto last_to_find = boundary.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_surface_id](const BoundaryFiniteElement& elem) {
				return elem.surface_id == t_surface_id;
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

void MeshLoader::Insert_Middle_Nodes()
{
	std::unordered_set <Edge> examined_edges;
	for (auto& elem : finite_elements) {
		std::vector<int> cur_nodes_id = elem.nodes_id;
		
		for (auto first_node = 0; first_node < 4; ++first_node)
			for (auto second_node = first_node + 1; second_node < 4; ++second_node) {
				Edge cur_edge(cur_nodes_id[first_node], cur_nodes_id[second_node]);
				auto found_edge = examined_edges.find(cur_edge);
				
				if (found_edge == examined_edges.end()) {
					Node new_node = Get_Middle_Node(cur_edge);
					nodes.push_back(new_node);
					cur_edge.UpdateMiddleNode(new_node.node_id);
					cur_nodes_id.push_back(new_node.node_id);
					
					examined_edges.insert(cur_edge);
				}
				else {
					elem.nodes_id.push_back(cur_edge.middle_node);
				}
			}
	}

	for (auto& ele_boundary : boundary) {
		std::vector<int> cur_nodes_id = ele_boundary.nodes_id;
		for (auto first_node = 0; first_node < 3; ++first_node)
			for (auto second_node = first_node + 1; second_node < 3; ++second_node) {
				Edge cur_edge(cur_nodes_id[first_node], cur_nodes_id[second_node]);
				auto found_edge = examined_edges.find(cur_edge);
				
				if (found_edge == examined_edges.end())
					found_edge = examined_edges.find({ cur_nodes_id[second_node], cur_nodes_id[first_node] });
				
				ele_boundary.nodes_id.push_back(found_edge->middle_node);
			}
	}	
}

std::vector<std::set<int>> MeshLoader::Get_Adjacent_Nodes() const
{
	std::vector<std::set<int>> res(nodes.size() + 1);

	for (const auto& cur_elem : finite_elements) {
		for (auto cur_node_id : cur_elem.nodes_id)
			for (auto adj_node_id : cur_elem.nodes_id)
				if (cur_node_id != adj_node_id)
					res[cur_node_id].insert(adj_node_id);
	}
	return res;
}


std::array<double, 3> MeshLoader::Get_Middle_Of_Edge(const Edge& t_edge)
{
	return
	{ (get<0>(nodes[t_edge.nodes.first - 1].coors) + get<0>(nodes[t_edge.nodes.second - 1].coors)) * 1 / 2
	, (get<1>(nodes[t_edge.nodes.first - 1].coors) + get<1>(nodes[t_edge.nodes.second - 1].coors)) * 1 / 2
	, (get<2>(nodes[t_edge.nodes.first - 1].coors) + get<2>(nodes[t_edge.nodes.second - 1].coors)) * 1 / 2 };
}

Node MeshLoader::Get_Middle_Node(const Edge& cur_edge)
{
	Node new_node(nodes.size() + 1, Get_Middle_Of_Edge(cur_edge), 0);
	return new_node;
}

Node& MeshLoader::Get_Node_By_ID(int t_id)
{
	return nodes.at(t_id);
}

std::array<double, 3> MeshLoader::Get_Coors_Node_By_ID(int t_id)
{
	return nodes.at(t_id).coors;
}



std::vector<FiniteElement> MeshLoader::Get_Nodes_By_Boundary(int boundary_id) const {
	std::vector<FiniteElement> res_finite_ele;
	std::set<int> res_nodes;
	auto first_boundary = boundary.begin();
	auto last_boundary = boundary.end();
	while (first_boundary != last_boundary) {
		auto found_boundary = std::find_if(first_boundary, last_boundary,
			[boundary_id](const BoundaryFiniteElement& cur_boundary) {
				return cur_boundary.surface_id == boundary_id;
			});

		if (found_boundary != last_boundary) {
			for (const auto& node_in_boundary : found_boundary->nodes_id) {
				res_nodes.insert(node_in_boundary);
			}
			first_boundary = found_boundary + 1;
		}
	}

	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();
	while (first_to_find != last_to_find) {
		auto found_finite_ele = std::find_if(first_to_find, last_to_find,
			[res_nodes](const FiniteElement& cur_finite_ele) {
				int check = 0;
				for (const auto& ele : res_nodes) {
					if (ele == cur_finite_ele.nodes_id.at(0)) ++check;
					if (ele == cur_finite_ele.nodes_id.at(1)) ++check;
					if (ele == cur_finite_ele.nodes_id.at(2)) ++check;
				}
				if (check == 3) return 1;
				return 0;
			});

		if (found_finite_ele != last_to_find) {
			res_finite_ele.push_back(*found_finite_ele);
			first_to_find = found_finite_ele + 1;
		}
	}
	return res_finite_ele;
}


double MeshLoader::Get_Area_By_Nodes(Node& a, Node& b, Node& c) {
	double x = sqrt(pow(a.coors[0] - b.coors[0], 2) + pow(a.coors[2] - b.coors[2], 2) + pow(a.coors[1] - b.coors[1], 2));
	double y = sqrt(pow(a.coors[0] - c.coors[0], 2) + pow(a.coors[2] - c.coors[2], 2) + pow(a.coors[1] - c.coors[1], 2));
	double z = sqrt(pow(b.coors[0] - c.coors[0], 2) + pow(b.coors[2] - c.coors[2], 2) + pow(b.coors[1] - c.coors[1], 2));
	double p = (x + y + z) / 2;
	return sqrt(p * (p - x) * (p - y) * (p - z));
}


double MeshLoader::Get_Area(const FiniteElement& finite_ele) {
	
	Node a = Get_Node_By_ID(finite_ele.nodes_id.at(0));
	Node b = Get_Node_By_ID(finite_ele.nodes_id.at(1));
	Node c = Get_Node_By_ID(finite_ele.nodes_id.at(2));
	return Get_Area_By_Nodes(a, b, c);
}


void MeshLoader::AreaTriangle(const MeshLoader& Mesh, int border_id, const std::string& name_file) {
	std::vector<FiniteElement> Container = Mesh.Get_Nodes_By_Boundary(border_id);
	std::ofstream ofstr(name_file + ".txt");
	if (!ofstr) throw NotFoundFiles();
	else {

		/*std::ranges::sort(Container, std::less<>{}, [&](const FiniteElement& ele) {
			return Get_Area(ele);
		});*/

		std::ranges::sort(Container, [&](const FiniteElement& fe1, const FiniteElement& fe2) {
			return Get_Area(fe1)>Get_Area(fe2);
			});
		
		auto transform = [&](const FiniteElement& finite_ele) {
			return finite_ele.nodes_id;
		};
		
		auto out = Container | std::ranges::views::drop(9) | std::ranges::views::take(10) | std::ranges::views::transform(transform);

		for (const auto& elem : out)
			std::ranges::copy(elem, std::ostream_iterator<int>(ofstr, "\n"));
	};
}












